{% extends '!layout.html' %}

{% block footer %}
<hr \>
<section class="footer">
<br><a target="_blank" href="http://www.nist.gov/public_affairs/privacy.cfm#privpolicy">Privacy Policy</a> | <a target="_blank" href="http://www.nist.gov/public_affairs/privacy.cfm#secnot">Security Notice</a> | <a href="http://www.nist.gov/public_affairs/privacy.cfm#accesstate">Accessibility Statement</a>
<br><em>United States government work, not subject to copyright in the United States</em>
</section>
{% endblock %}

{%- block header %}
<header class="nist-header">
  <meta charset="utf-8">
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <title>NIST Pages Template</title>
  <meta name="viewport" content="width=device-width">
  <meta name="description" content="{{ docstitle }}">
  <link rel="canonical" href="{{ pathto('', 1) }}">
  <link rel="shortcut icon" href="/favicon.ico">

  <!-- Google Analytics -->
  <!--<script type="text/javascript" id="_fed_an_js_tag" src="http://www.nist.gov/js/federated-analytics.all.min.js?agency=NIST&subagency=mml&pua=UA-66610693-1&yt=true&exts=ppsx,pps,f90,sch,rtf,wrl,txz,m1v,xlsm,msi,xsd,f,tif,eps,mpg,xml,pl,xlt,c"></script> -->
  <!-- DAP Analytics -->
  <script type="text/javascript" id="_fed_an_ua_tag" src="https://dap.digitalgov.gov/Universal-Federated-Analytics-Min.js?agency=DOC&subagefncy=NIST&pua=UA-66610693-1&yt=true&exts=ppsx,pps,f90,sch,rtf,wrl,txz,m1v,xlsm,msi,xsd,f,tif,eps,mpg,xml,pl,xlt,c"></script>

  <!-- Custom CSS -->
  <link rel="stylesheet" href="{{ pathto('', 1) }}/_static/NISTStyle.css">
  <link rel="stylesheet" href="{{ pathto('', 1) }}/_static/NISTPages.css">

  <h1>
    <a class="nist-logo" target="_blank" href="http://www.nist.gov/" title="Go to nist.gov">National Institute of Standards and Technology</a>
  </h1>
  <div class="nist-links">
    <a class="nist-links-button" target="_blank" href="http://www.nist.gov">NIST Website</a>
    <a class="nist-links-button mobile-hide" target="_blank" href="http://www.nist.gov/public_affairs/nandyou.cfm">About NIST</a>
    <a class="nist-links-button mobile-hide" target="_blank" href="https://github.com/usnistgov">usnistgov on GitHub</a>
  </div>
  
  <!-- HTML5 shim and Respond.js IE8 support of HTML5 elements and media queries -->
  <!--[if lt IE 9]>
  <script src="https://oss.maxcdn.com/libs/html5shiv/3.7.0/html5shiv.js"></script>
    <script src="http://html5shim.googlecode.com/svn/trunk/html5.js"></script>
  <script src="https://oss.maxcdn.com/libs/respond.js/1.4.2/respond.min.js"></script>
  <![endif]-->
</header>
{% endblock %}

